"""Project: PhiK - correlation analyzer library

Created: 2018/09/05

Description:
    Functions for doing the significance evaluation of an hypothesis test of variable independence
    using a contingency table.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""
from typing import Tuple, Union

import numpy as np
import pandas as pd
import math
import itertools
import warnings

from scipy import stats
from scipy import special, optimize

from phik import definitions as defs
from .binning import bin_data, create_correlation_overview_table
from .statistics import get_chi2_using_dependent_frequency_estimates
from .statistics import estimate_ndof, theoretical_ndof
from .simulation import sim_chi2_distribution
from .data_quality import dq_check_nunique_values, dq_check_hist2d
from .utils import array_like_to_dataframe, guess_interval_cols


def fit_test_statistic_distribution(
    chi2s: Union[list, np.ndarray], nbins: int = 50
) -> Tuple[float, float, float, float]:
    """
    Fit the hybrid chi2-distribution to the data to find f.

    Perform a binned likelihood fit to the data to find the optimal value for the fraction f in
    h(x|f) = N * (f * chi2(x, ndof) + (1-f) * gauss(x, ndof, sqrt(ndof))
    The parameter ndof is fixed in the fit using ndof = mean(x). The total number of datapoints N is also fixed.

    :param list chi2s: input data - a list of chi2 values
    :param int nbins: in order to fit the data a histogram is created with nbins number of bins
    :returns: f, ndof, sigma (width of gauss), bw (bin width)
    """

    def myfunc(x, N, f, k, sigma):
        return N * (f * stats.chi2.pdf(x, k) + (1 - f) * stats.norm.pdf(x, k, sigma))

    ffunc = lambda x, f: myfunc(x, nsim * bw, f, kmean, lsigma)

    def gtest(p, x, y):
        f = ffunc(x, *p)
        ll = f - special.xlogy(y, f) + special.gammaln(y + 1)
        return np.sqrt(ll)

    kmean = np.mean(chi2s)  # NOTE: this is used to fix kmean in the fit!
    lsigma = np.sqrt(kmean)  # NOTE: this is used to fix sigma in the fit!
    nsim = len(chi2s)  # NOTE: this is used to fix N in fit (N=nsim*bw) !

    yhist, xbounds = np.histogram(chi2s, bins=nbins)
    bw = xbounds[1] - xbounds[0]  # NOTE: this is used to fix N in fit (N=nsim*bw) !
    xhist = xbounds[:-1] + np.diff(xbounds) / 2

    initGuess = (1.0,)  # starting value for parameter f
    res = optimize.least_squares(
        gtest, initGuess, bounds=((0.0,), (1,)), args=(xhist, yhist)
    )

    return res.x[0], kmean, lsigma, bw


def hfunc(x: float, N: float, f: float, k: float, sigma: float) -> float:
    """
    Definition of the combined probability density function h(x)

    h(x|f) = N * (f * chi2(x, k) + (1-f) * gauss(x, k, sigma)

    :param float x: x
    :param float N: normalisation
    :param float f: fraction [0,1]
    :param float k: ndof of chi2 function and mean of gauss
    :param float sigma: width of gauss
    :return: h(x|f)
    """
    return N * (f * stats.chi2.pdf(x, k) + (1 - f) * stats.norm.pdf(x, k, sigma))


def significance_from_chi2_ndof(chi2: float, ndof: float) -> Tuple[float, float]:
    """
    Convert a chi2 into significance using knowledge about the number of degrees of freedom

    Conversion is done using asymptotic approximation.

    :param float chi2: chi2 value
    :param float ndof: number of degrees of freedom
    :returns: p_value, significance
    """
    p_value = stats.chi2.sf(chi2, ndof)
    z_value = -stats.norm.ppf(p_value)

    # scenario where p_value is too small to evaluate Z
    # use Chernoff approximation for p-value upper bound
    # see: https://en.wikipedia.org/wiki/Chi-squared_distribution
    if p_value == 0:
        z = chi2 / ndof
        u = -math.log(2 * math.pi) - ndof * math.log(z) + ndof * (z - 1)
        z_value = math.sqrt(u - math.log(u))

    return p_value, z_value


def significance_from_chi2_asymptotic(
    values: np.ndarray, chi2: float
) -> Tuple[float, float]:
    """
    Convert a chi2 into significance using knowledge about the number of degrees of freedom

    Convention is done using asymptotic approximation.

    :param float chi2: chi2 value
    :param float ndof: number of degrees of freedom
    :returns: p_value, significance
    """

    ndof = theoretical_ndof(values)
    p_value, z_value = significance_from_chi2_ndof(chi2, ndof)

    return p_value, z_value


def significance_from_chi2_MC(
    chi2: float,
    values: np.ndarray,
    nsim: int = 1000,
    lambda_: str = "log-likelihood",
    simulation_method: str = "multinominal",
    chi2s=None,
    njobs: int = -1,
) -> Tuple[float, float]:
    """
    Convert a chi2 into significance using knowledge about the shape of the chi2 distribution of simulated data

    Calculate significance based on simulation (MC method), using a simple percentile.

    :param float chi2: chi2 value
    :param list chi2s: provide your own chi2s values (optional)
    :param int njobs: number of parallel jobs used for simulation. default is -1. 1 uses no parallel jobs.
    :returns: pvalue, significance
    """

    # determine effective number of degrees of freedom using simulation
    if chi2s is None:
        chi2s = sim_chi2_distribution(
            values,
            nsim=nsim,
            lambda_=lambda_,
            simulation_method=simulation_method,
            njobs=njobs,
        )

    # calculate p_value based on simulation (MC method)
    empirical_p_value = 1.0 - stats.percentileofscore(chi2s, chi2) / 100.0
    empirical_z_value = -stats.norm.ppf(empirical_p_value)

    return empirical_p_value, empirical_z_value


def significance_from_chi2_hybrid(
    chi2: float,
    values: np.ndarray,
    nsim: int = 1000,
    lambda_: str = "log-likelihood",
    simulation_method: str = "multinominal",
    chi2s=None,
    njobs: int = -1,
) -> Tuple[float, float]:
    """
    Convert a chi2 into significance using a hybrid method

    This method combines the asymptotic method with the MC method, but applies several corrections:

    * use effective number of degrees of freedom instead of number of degrees of freedom. The effective number of\
      degrees of freedom is measured as mean(chi2s), with chi2s a list of simulated chi2 values.
    * for low statistics data sets, with on average less than 4 data points per bin, the distribution of chi2-values\
      is better described by h(x|f) then by the usual chi2-distribution. Use h(x|f) to convert the chi2 value to \
      the pvalue and significance.

    h(x|f) = N * (f * chi2(x, ndof) + (1-f) * gauss(x, ndof, sqrt(ndof))

    :param float chi2: chi2 value
    :param list chi2s: provide your own chi2s values (optional)
    :param float avg_per_bin: average number of data points per bin
    :param int njobs: number of parallel jobs used for simulation. default is -1. 1 uses no parallel jobs.
    :returns: p_value, significance
    """

    # determine effective number of degrees of freedom using simulation
    if chi2s is None:
        chi2s = sim_chi2_distribution(
            values,
            nsim=nsim,
            lambda_=lambda_,
            simulation_method=simulation_method,
            njobs=njobs,
        )

    # average number of records per bin
    avg_per_bin = values.sum() / values.shape[0] * values.shape[1]

    if avg_per_bin <= 4:
        f, endof, lsigma, bw = fit_test_statistic_distribution(chi2s)
        pvalue_h = f * stats.chi2.sf(chi2, endof) + (1 - f) * stats.norm.sf(
            chi2, endof, lsigma
        )
    else:
        endof = estimate_ndof(chi2s)
        pvalue_h = stats.chi2.sf(chi2, endof)

    zvalue_h = -stats.norm.ppf(pvalue_h)

    if pvalue_h == 0:
        # apply Chernoff approximation as upper bound for p-value
        # see: https://en.wikipedia.org/wiki/Chi-squared_distribution
        z = chi2 / endof
        u = -math.log(2 * math.pi) - endof * math.log(z) + endof * (z - 1)
        if avg_per_bin <= 4:
            u += -2 * math.log(f)
        zvalue_h = math.sqrt(u - math.log(u))

    return pvalue_h, zvalue_h


def significance_from_hist2d(
    values: np.ndarray,
    nsim: int = 1000,
    lambda_: str = "log-likelihood",
    simulation_method: str = "multinominal",
    significance_method: str = "hybrid",
    njobs: int = -1,
) -> Tuple[float, float]:
    """
    Calculate the significance of correlation of two variables based on the contingency table

    :param values: contingency table
    :param int nsim: number of simulations
    :param str lambda_: test statistic. Available options are [pearson, log-likelihood]
    :param str simulation_method: simulation method. Options: [multinominal, row_product_multinominal, \
     col_product_multinominal, hypergeometric].
    :param str significance_method: significance_method. Options: [asymptotic, MC, hybrid]
    :param int njobs: number of parallel jobs used for simulation. default is -1. 1 uses no parallel jobs.
    :return: pvalue, significance
    """

    # chi2 of the data
    chi2 = get_chi2_using_dependent_frequency_estimates(values, lambda_=lambda_)

    if significance_method == "asymptotic":
        # calculate pvalue and zvalue based on chi2 and ndof (asymptotic method)
        pvalue, zvalue = significance_from_chi2_asymptotic(values, chi2)
    elif significance_method == "MC":
        # calculate pvalue based on simulation (MC method)
        pvalue, zvalue = significance_from_chi2_MC(
            chi2,
            values,
            nsim=nsim,
            lambda_=lambda_,
            simulation_method=simulation_method,
            njobs=njobs,
        )
    elif significance_method == "hybrid":
        # low statistics : calculate pvalue and zvalue using h(x|f) and endof
        # high statistics: calculate pvalue and zvalue using chi2-distribution and endof
        pvalue, zvalue = significance_from_chi2_hybrid(
            chi2,
            values,
            nsim=nsim,
            lambda_=lambda_,
            simulation_method=simulation_method,
            njobs=njobs,
        )
    else:
        raise NotImplementedError(
            "simulation_method {0:s} is unknown".format(simulation_method)
        )

    return pvalue, zvalue


def significance_from_rebinned_df(
    data_binned: pd.DataFrame,
    lambda_: str = "log-likelihood",
    simulation_method: str = "multinominal",
    nsim: int = 1000,
    significance_method: str = "hybrid",
    dropna: bool = True,
    drop_underflow: bool = True,
    drop_overflow: bool = True,
    njobs: int = -1,
) -> pd.DataFrame:
    """
    Calculate significance of correlation of all variable combinations in the DataFrame

    :param data_binned: input binned DataFrame
    :param int nsim: number of simulations
    :param str lambda_: test statistic. Available options are [pearson, log-likelihood]
    :param str simulation_method: simulation method. Options: [mutlinominal, row_product_multinominal, \
     col_product_multinominal, hypergeometric].
    :param str significance_method: significance_method. Options: [asymptotic, MC, hybrid]
    :param bool dropna: remove NaN values with True
    :param bool drop_underflow: do not take into account records in underflow bin when True (relevant when binning\
    a numeric variable)
    :param bool drop_overflow: do not take into account records in overflow bin when True (relevant when binning\
    a numeric variable)
    :param int njobs: number of parallel jobs used for simulation. default is -1.
    :return: significance matrix
    """

    if not dropna:
        # if not dropna replace the NaN values with the string NaN. Otherwise the rows with NaN are dropped
        # by the groupby.
        data_binned.replace(np.nan, defs.NaN, inplace=True)
    if drop_underflow:
        data_binned.replace(defs.UF, np.nan, inplace=True)
    if drop_overflow:
        data_binned.replace(defs.OF, np.nan, inplace=True)

    # cache column order (https://github.com/KaveIO/PhiK/issues/1)
    column_order = data_binned.columns
    signifs = []
    for i, (c0, c1) in enumerate(
        itertools.combinations_with_replacement(data_binned.columns.values, 2)
    ):
        datahist = (
            data_binned.groupby([c0, c1])[c0].count().to_frame().unstack().fillna(0)
        )
        if 1 in datahist.shape or 0 in datahist.shape:
            signifs.append((c0, c1, np.nan))
            warnings.warn(
                "Too few unique values for variable {0:s} ({1:d}) or {2:s} ({3:d}) to calculate significance".format(
                    c0, datahist.shape[0], c1, datahist.shape[1]
                )
            )
            continue

        datahist.columns = datahist.columns.droplevel()
        datahist = datahist.values
        pvalue, zvalue = significance_from_hist2d(
            datahist,
            nsim=nsim,
            lambda_=lambda_,
            simulation_method=simulation_method,
            significance_method=significance_method,
            njobs=njobs,
        )
        signifs.append((c0, c1, zvalue))

    if len(signifs) == 0:
        return pd.DataFrame(np.nan, index=column_order, columns=column_order)

    significance_overview = create_correlation_overview_table(signifs)

    # restore column order
    significance_overview = significance_overview.reindex(columns=column_order)
    significance_overview = significance_overview.reindex(index=column_order)

    return significance_overview


def significance_matrix(
    df: pd.DataFrame,
    interval_cols: list = None,
    lambda_: str = "log-likelihood",
    simulation_method: str = "multinominal",
    nsim: int = 1000,
    significance_method: str = "hybrid",
    bins: Union[int, list, np.ndarray, dict] = 10,
    dropna: bool = True,
    drop_underflow: bool = True,
    drop_overflow: bool = True,
    verbose: bool = True,
    njobs: int = -1,
) -> pd.DataFrame:
    """
    Calculate significance of correlation of all variable combinations in the dataframe

    :param pd.DataFrame df: input data
    :param list interval_cols: column names of columns with interval variables.
    :param int nsim: number of simulations
    :param str lambda_: test statistic. Available options are [pearson, log-likelihood]
    :param str simulation_method: simulation method. Options: [mutlinominal, row_product_multinominal, \
     col_product_multinominal, hypergeometric].
    :param int nsim: number of simulated datasets
    :param str significance_method: significance_method. Options: [asymptotic, MC, hybrid]
        :param bool dropna: remove NaN values with True
    :param bins: number of bins, or a list of bin edges (same for all columns), or a dictionary where per column the bins are specified. (default=10)\
    E.g.: bins = {'mileage':5, 'driver_age':[18,25,35,45,55,65,125]}
    :param bool dropna: remove NaN values with True
    :param bool drop_underflow: do not take into account records in underflow bin when True (relevant when binning\
    a numeric variable)
    :param bool drop_overflow: do not take into account records in overflow bin when True (relevant when binning\
    a numeric variable)
    :param bool verbose: if False, do not print all interval columns that are guessed
    :param int njobs: number of parallel jobs used for simulation. default is -1. 1 uses no parallel jobs.
    :return: significance matrix
    """

    if interval_cols is None:
        interval_cols = guess_interval_cols(df, verbose)

    df_clean, interval_cols_clean = dq_check_nunique_values(
        df, interval_cols, dropna=dropna
    )

    data_binned = bin_data(df_clean, interval_cols_clean, bins=bins)

    return significance_from_rebinned_df(
        data_binned,
        lambda_=lambda_,
        simulation_method=simulation_method,
        nsim=nsim,
        significance_method=significance_method,
        dropna=dropna,
        drop_underflow=drop_underflow,
        drop_overflow=drop_overflow,
        njobs=njobs,
    )


def significance_from_array(
    x: Union[np.ndarray, pd.Series],
    y: Union[np.ndarray, pd.Series],
    num_vars=None,
    bins: Union[int, list, np.ndarray, dict] = 10,
    quantile: bool = False,
    lambda_: str = "log-likelihood",
    nsim: int = 1000,
    significance_method: str = "hybrid",
    simulation_method: str = "multinominal",
    dropna: bool = True,
    drop_underflow: bool = True,
    drop_overflow: bool = True,
    njobs: int = -1,
) -> Tuple[float, float]:
    """
    Calculate the significance of correlation

    Calculate the significance of correlation for two variables which can be of interval, oridnal or categorical type.\
    Interval variables will be binned.

    :param x: array-like input
    :param y: array-like input
    :param num_vars: list of numeric variables which need to be binned, e.g. ['x'] or ['x','y']
    :param bins: number of bins, or a list of bin edges (same for all columns), or a dictionary where per column the bins are specified. (default=10)\
    E.g.: bins = {'mileage':5, 'driver_age':[18,25,35,45,55,65,125]}
    :param quantile: when bins is an integer, uniform bins (False) or bins based on quantiles (True)
    :param str lambda_: test statistic. Available options are [pearson, log-likelihood]
    :param int nsim: number of simulated datasets
    :param str simulation_method: simulation method. Options: [mutlinominal, row_product_multinominal, \
    col_product_multinominal, hypergeometric].
    :param str significance_method: significance_method. Options: [asymptotic, MC, hybrid]
    :param bool dropna: remove NaN values with True
    :param bool drop_underflow: do not take into account records in underflow bin when True (relevant when binning\
    a numeric variable)
    :param bool drop_overflow: do not take into account records in overflow bin when True (relevant when binning\
    a numeric variable)
    :param int njobs: number of parallel jobs used for simulation. default is -1. 1 uses no parallel jobs.
    :return: p-value, significance
    """
    if num_vars is None:
        num_vars = []
    elif isinstance(num_vars, str):
        num_vars = [num_vars]

    if len(num_vars) > 0:
        df = array_like_to_dataframe(x, y)
        x, y = bin_data(df, num_vars, bins=bins, quantile=quantile).T.values

    return significance_from_binned_array(
        x,
        y,
        lambda_=lambda_,
        significance_method=significance_method,
        nsim=nsim,
        simulation_method=simulation_method,
        dropna=dropna,
        drop_underflow=drop_underflow,
        drop_overflow=drop_overflow,
        njobs=njobs,
    )


def significance_from_binned_array(
    x: Union[np.ndarray, pd.Series],
    y: Union[np.ndarray, pd.Series],
    lambda_: str = "log-likelihood",
    significance_method: str = "hybrid",
    nsim: int = 1000,
    simulation_method: str = "multinominal",
    dropna: bool = True,
    drop_underflow: bool = True,
    drop_overflow: bool = True,
    njobs: int = -1,
) -> Tuple[float, float]:
    """
    Calculate the significance of correlation

    Calculate the significance of correlation for two variables which can be of interval, oridnal or categorical type. \
    Interval variables need to be binned.

    :param x: array-like input
    :param y: array-like input
    :param str lambda_: test statistic. Available options are [pearson, log-likelihood]
    :param str simulation_method: simulation method. Options: [multinominal, row_product_multinominal, \
    col_product_multinominal, hypergeometric].
    :param int nsim: number of simulated datasets
    :param str significance_method: significance_method. Options: [asymptotic, MC, hybrid]
    :param bool dropna: remove NaN values with True
    :param bool drop_underflow: do not take into account records in underflow bin when True (relevant when binning\
    a numeric variable)
    :param bool drop_overflow: do not take into account records in overflow bin when True (relevant when binning\
    a numeric variable)
    :param int njobs: number of parallel jobs used for simulation. default is -1. 1 uses no parallel jobs.
    :return: p-value, significance
    """

    if not dropna:
        x = pd.Series(x).fillna(defs.NaN).astype(str).values
        y = (
            pd.Series(y).fillna(defs.NaN).astype(str).values
        )  # crosstab cannot handle mixed type y!

    if drop_underflow or drop_overflow:
        x = pd.Series(x).astype(str).values
        y = pd.Series(y).astype(str).values
        if drop_underflow:
            x[np.where(x == defs.UF)] = np.nan
            y[np.where(y == defs.UF)] = np.nan
        if drop_overflow:
            y[np.where(y == defs.OF)] = np.nan
            x[np.where(x == defs.OF)] = np.nan

    hist2d = pd.crosstab(x, y).values

    if not dq_check_hist2d(hist2d):
        return np.nan, np.nan

    return significance_from_hist2d(
        hist2d,
        lambda_=lambda_,
        significance_method=significance_method,
        simulation_method=simulation_method,
        nsim=nsim,
        njobs=njobs,
    )



# flake8: noqa
from phik.version import version as __version__

# pandas dataframe decorators
from phik import decorators

# array functions
from phik.phik import phik_from_array
from phik.significance import significance_from_array
from phik.outliers import outlier_significance_from_array

# dataframe functions
from phik.phik import phik_matrix, global_phik_array
from phik.significance import significance_matrix
from phik.outliers import outlier_significance_matrices, outlier_significance_matrix



from typing import Union

import pandas as pd
import numpy as np


def array_like_to_dataframe(
    x: Union[pd.Series, list, np.ndarray], y: Union[pd.Series, list, np.ndarray]
) -> pd.DataFrame:
    """Concat two array-like data structures into a DataFrame

    :param x: pd.Series, list or np.ndarray
    :param y: pd.Series, list or np.ndarray
    :return: pd.DataFrame
    """
    x_name = x.name if isinstance(x, pd.Series) else "x"
    y_name = y.name if isinstance(y, pd.Series) else "y"

    return pd.DataFrame(np.array([x, y]).T, columns=[x_name, y_name])


def guess_interval_cols(df: pd.DataFrame, verbose: bool = False) -> list:
    """Select columns that have a dtype part of np.number

    :param df: DataFrame
    :param bool verbose: print all interval columns that are guessed
    :return: list of interval columns
    """
    interval_cols = df.select_dtypes(include=[np.number]).columns.tolist()
    if interval_cols and verbose:
        print("interval columns not set, guessing: {}".format(str(interval_cols)))

    if not isinstance(interval_cols, list):
        raise ValueError("Could not guess interval columns")
    return interval_cols


def make_shapes_equal(observed: pd.DataFrame, expected: pd.DataFrame) -> pd.DataFrame:
    """Make observed and expected (pandas) histograms equal in shape

    Sometimes expected histogram shape need filling / pruning to make its shape equal to observed. Give expected the
    same number of columns and rows. Needed for proper histogram comparison.

    :param pd.DataFrame observed: The observed contingency table. The table contains the observed number of occurrences in each cell.
    :param pd.DataFrame expected: The expected contingency table. The table contains the expected number of occurrences in each cell.
    :return: expected frequencies, having the same shape as observed
    """
    # columns
    o_cols = observed.columns.tolist()
    e_cols = expected.columns.tolist()
    o_cols_missing = list(set(e_cols) - set(o_cols))
    e_cols_missing = list(set(o_cols) - set(e_cols))
    # index
    o_idx = observed.index.tolist()
    e_idx = expected.index.tolist()
    o_idx_missing = list(set(e_idx) - set(o_idx))
    e_idx_missing = list(set(o_idx) - set(e_idx))

    # make expected columns equal to observed
    for c in o_cols_missing:
        observed[c] = 0.0
    for c in e_cols_missing:
        expected[c] = 0.0
    observed.columns = sorted(observed.columns)
    expected.columns = sorted(expected.columns)
    # this should always be a match now
    assert len(observed.columns) == len(expected.columns)

    # make expected index equal to observed
    for i in o_idx_missing:
        observed.loc[i] = np.zeros(len(observed.columns))
    for i in e_idx_missing:
        expected.loc[i] = np.zeros(len(expected.columns))
    # this should always be a match now
    assert len(observed.index) == len(expected.index)

    return expected



"""Project: PhiK - correlation analyzer library

Created: 2018/09/06

Description:
    Functions to create nice correlation overview and matrix plots

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""
from typing import Tuple, Union, Callable, Dict

import os
import itertools
import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib import colors

from .binning import bin_data
from .phik import phik_from_rebinned_df, global_phik_from_rebinned_df
from .significance import significance_from_rebinned_df
from .outliers import outlier_significance_matrix_from_rebinned_df
from .data_quality import dq_check_nunique_values
from .utils import guess_interval_cols


def plot_hist_and_func(
    data: Union[list, np.ndarray, pd.Series],
    func: Callable,
    funcparams,
    xbins=False,
    labels=None,
    xlabel="",
    ylabel="",
    title="",
    xlimit=None,
    alpha=1,
):
    """
    Create a histogram of the provided data and overlay with a function.

    :param list data: data
    :param function func: function of the type f(x, a, b, c) where parameters a, b, c are optional
    :param list funcparams: parameter values to be given to the function, to be specified as [a, b, c]
    :param xbins: specify binning of histogram, either by giving the number of bins or a list of bin edges
    :param labels: labels of histogram and function to be used in the legend
    :param xlabel: figure xlabel
    :param ylabel: figure ylabel
    :param title: figure title
    :param xlimit: x limits figure
    :param alpha: alpha histogram
    :return:
    """
    if labels is None:
        labels = ["", ""]

    # If binning is not specified, create binning here
    if not np.any(xbins) and not xlimit:
        xmin = np.min(data)
        xmax = np.max(data)
        xnbins = int(len(data) / 50 + 1)
        xbins = np.linspace(xmin, xmax, xnbins)
    elif type(xbins) == int or type(xbins) == float:
        xmin = np.min(data)
        if xlimit:
            xmin = xlimit[0]
        xmax = np.max(data)
        if xmax:
            xmax = xlimit[1]
        xnbins = int(xbins + 1)
        xbins = np.linspace(xmin, xmax, xnbins)

    # Plot a histogram of the data
    plt.hist(data, bins=xbins, label=labels[0], alpha=alpha)

    # Find bin centers for plotting the function
    xvals = xbins[:-1] + np.diff(xbins)[0] / 2
    bw = xbins[1] - xbins[0]
    # Plot the fit
    plt.plot(
        xvals, len(data) * bw * func(xvals, *funcparams), linewidth=2, label=labels[1]
    )

    if xlabel:
        plt.xlabel(xlabel)
    if ylabel:
        plt.ylabel(ylabel)
    if title:
        plt.title(title)

    if len(labels[0]) > 0:
        plt.legend()


def plot_correlation_matrix(
    matrix_colors: np.ndarray,
    x_labels: list,
    y_labels: list,
    pdf_file_name: str = "",
    title: str = "correlation",
    vmin: float = -1,
    vmax: float = 1,
    color_map: str = "RdYlGn",
    x_label: str = "",
    y_label: str = "",
    top: int = 20,
    matrix_numbers: np.ndarray = None,
    print_both_numbers: bool = True,
    figsize: tuple = (7, 5),
    usetex: bool = False,
    identity_layout: bool = True,
    fontsize_factor: float = 1,
) -> None:
    """Create and plot correlation matrix.

    Copied with permission from the eskapade package (pip install eskapade)

    :param matrix_colors: input correlation matrix
    :param list x_labels: Labels for histogram x-axis bins
    :param list y_labels: Labels for histogram y-axis bins
    :param str pdf_file_name: if set, will store the plot in a pdf file
    :param str title: if set, title of the plot
    :param float vmin: minimum value of color legend (default is -1)
    :param float vmax: maximum value of color legend (default is +1)
    :param str x_label: Label for histogram x-axis
    :param str y_label: Label for histogram y-axis
    :param str color_map: color map passed to matplotlib pcolormesh. (default is 'RdYlGn')
    :param int top: only print the top 20 characters of x-labels and y-labels. (default is 20)
    :param matrix_numbers: input matrix used for plotting numbers. (default it matrix_colors)
    :param identity_layout: Plot diagonal from right top to bottom left (True) or bottom left to top right (False)
    """
    if not isinstance(matrix_colors, np.ndarray):
        raise TypeError("matrix_colors is not a numpy array.")

    # basic matrix checks
    assert (matrix_colors.shape[0] == len(y_labels)) or (
        matrix_colors.shape[0] + 1 == len(y_labels)
    ), "matrix_colors shape inconsistent with number of y-labels"
    assert (matrix_colors.shape[1] == len(x_labels)) or (
        matrix_colors.shape[1] + 1 == len(x_labels)
    ), "matrix_colors shape inconsistent with number of x-labels"
    if matrix_numbers is None:
        matrix_numbers = matrix_colors
        print_both_numbers = False  # only one set of numbers possible
    else:
        assert matrix_numbers.shape[0] == len(
            y_labels
        ), "matrix_numbers shape inconsistent with number of y-labels"
        assert matrix_numbers.shape[1] == len(
            x_labels
        ), "matrix_numbers shape inconsistent with number of x-labels"

    if identity_layout:
        matrix_colors = np.array([a[::-1] for a in matrix_colors])
        x_labels = x_labels[::-1]
        if matrix_numbers is not None:
            matrix_numbers = np.array([a[::-1] for a in matrix_numbers])

    plt.rc("text", usetex=usetex)

    fig, ax = plt.subplots(figsize=figsize)
    # cmap = 'RdYlGn' #'YlGn'
    norm = colors.Normalize(vmin=vmin, vmax=vmax)
    img = ax.pcolormesh(
        matrix_colors, cmap=color_map, edgecolor="w", linewidth=1, norm=norm
    )

    # set x-axis properties
    def tick(lab):
        """Get tick."""
        if isinstance(lab, (float, int)):
            lab = "NaN" if np.isnan(lab) else "{0:.0f}".format(lab)
        lab = str(lab)
        if len(lab) > top:
            lab = lab[:17] + "..."
        return lab

    # reduce default fontsizes in case too many labels?
    # nlabs = max(len(y_labels), len(x_labels))

    # axis ticks and tick labels
    if len(x_labels) == matrix_colors.shape[1] + 1:
        ax.set_xticks(np.arange(len(x_labels)))
    else:
        ax.set_xticks(np.arange(len(x_labels)) + 0.5)
    ax.set_xticklabels(
        [tick(lab) for lab in x_labels],
        rotation="vertical",
        fontsize=10 * fontsize_factor,
    )

    if len(y_labels) == matrix_colors.shape[0] + 1:
        ax.set_yticks(np.arange(len(y_labels)))
    else:
        ax.set_yticks(np.arange(len(y_labels)) + 0.5)
    ax.set_yticklabels(
        [tick(lab) for lab in y_labels],
        rotation="horizontal",
        fontsize=10 * fontsize_factor,
    )

    # Turn ticks off in case no labels are provided
    if len(x_labels) == 1 and len(x_labels[0]) == 0:
        plt.tick_params(
            axis="x",  # changes apply to the x-axis
            which="both",  # both major and minor ticks are affected
            bottom=False,  # ticks along the bottom edge are off
            top=False,  # ticks along the top edge are off
            labelbottom=False,
        )
    if len(y_labels) == 1 and len(y_labels[0]) == 0:
        plt.tick_params(
            axis="y",  # changes apply to the x-axis
            which="both",  # both major and minor ticks are affected
            left=False,  # ticks along the bottom edge are off
            right=False,  # ticks along the top edge are off
            labelbottom=False,
        )

    # make plot look pretty
    ax.set_title(title, fontsize=14 * fontsize_factor)
    if x_label:
        ax.set_xlabel(x_label, fontsize=12 * fontsize_factor)
    if y_label:
        ax.set_ylabel(y_label, fontsize=12 * fontsize_factor)

    fig.colorbar(img)

    # annotate with correlation values
    numbers_set = (
        [matrix_numbers] if not print_both_numbers else [matrix_numbers, matrix_colors]
    )
    for i in range(matrix_numbers.shape[1]):
        for j in range(matrix_numbers.shape[0]):
            point_color = float(matrix_colors[j][i])
            white_cond = (
                (point_color < 0.7 * vmin)
                or (point_color >= 0.7 * vmax)
                or np.isnan(point_color)
            )
            y_offset = 0.5
            for m, matrix in enumerate(numbers_set):
                if print_both_numbers:
                    if m == 0:
                        y_offset = 0.7
                    elif m == 1:
                        y_offset = 0.25
                point = float(matrix[j][i])
                label = "NaN" if np.isnan(point) else "{0:.2f}".format(point)
                color = "w" if white_cond else "k"
                ax.annotate(
                    label,
                    xy=(i + 0.5, j + y_offset),
                    color=color,
                    horizontalalignment="center",
                    verticalalignment="center",
                    fontsize=10 * fontsize_factor,
                )

    plt.tight_layout()

    # save plot in file
    if pdf_file_name:
        pdf_file = PdfPages(pdf_file_name)
        plt.savefig(pdf_file, format="pdf", bbox_inches="tight", pad_inches=0)
        plt.close()
        pdf_file.close()


def correlation_report(
    data: pd.DataFrame,
    interval_cols: list = None,
    bins=10,
    quantile: bool = False,
    do_outliers: bool = True,
    pdf_file_name: str = "",
    significance_threshold: float = 3,
    correlation_threshold: float = 0.5,
    noise_correction: bool = True,
    store_each_plot: bool = False,
    lambda_significance: str = "log-likelihood",
    simulation_method: str = "multinominal",
    nsim_chi2: int = 1000,
    significance_method: str = "asymptotic",
    CI_method: str = "poisson",
    verbose: bool = True,
) -> Tuple[pd.DataFrame, pd.DataFrame, Dict[str, pd.DataFrame], Dict[str, str]]:
    """
    Create a correlation report for the given dataset.

    The following quantities are calculated:

    * The phik correlation matrix
    * The significance matrix
    * The outlier significances measured in pairs of variables. (optional)

    :param data: input dataframe
    :param interval_cols: list of columns names of columns containing interval data
    :param bins: number of bins, or a list of bin edges (same for all columns), or a dictionary where per column the bins are specified. (default=10)\
    E.g.: bins = {'mileage':5, 'driver_age':[18,25,35,45,55,65,125]}
    :param quantile: when bins is an integer, uniform bins (False) or bins based on quantiles (True)
    :param do_outliers: Evaluate outlier significances of variable pairs (when True)
    :param pdf_file_name: file name of the pdf where the results are stored
    :param store_each_plot: store each plot in folder derived from pdf_file_name. If true, single pdf is no longer stored. Default is false.
    :param significance_threshold: evaluate outlier significance for all variable pairs with a significance of \
     uncorrelation higher than this threshold
    :param correlation_threshold: evaluate outlier significance for all variable pairs with a phik correlation \
     higher than this threshold
    :param noise_correction: Apply noise correction in phik calculation
    :param lambda_significance: test statistic used in significance calculation. Options: [pearson, log-likelihood]
    :param simulation_method: sampling method using in significance calculation. Options: [mutlinominal, \
    row_product_multinominal, col_product_multinominal, hypergeometric]
    :param nsim_chi2: number of simulated datasets in significance calculation.
    :param significance_method: method for significance calculation. Options: [asymptotic, MC, hybrid]
    :param CI_method: method for uncertainty calculation for outlier significance calculation. Options: [poisson, \
    exact_poisson]
    :param bool verbose: if False, do not print all interval columns that are guessed
    :returns: phik_matrix (pd.DataFrame), global_phik (np.array), significance_matrix (pd.DataFrame), \
    outliers_overview (dictionary), output_files (dictionary)
    """

    if interval_cols is None:
        interval_cols = guess_interval_cols(data, verbose)

    data_clean, interval_cols_clean = dq_check_nunique_values(data, interval_cols)

    # create pdf(s) to save plots
    output_files = dict()
    plot_file_name = ""
    if store_each_plot:
        folder = os.path.dirname(pdf_file_name)
        folder += "/" if folder else "./"
        # if each plot is stored, single overview file is no longer stored.
        # (b/c of problem with multiple PdfPages)
        pdf_file_name = ""
    if pdf_file_name:
        pdf_file = PdfPages(pdf_file_name)

    data_binned, binning_dict = bin_data(
        data_clean, interval_cols_clean, bins=bins, quantile=quantile, retbins=True
    )

    ### 1. Phik
    if store_each_plot:
        plot_file_name = folder + "phik_matrix.pdf"
        output_files["phik_matrix"] = plot_file_name
    phik_matrix = phik_from_rebinned_df(data_binned, noise_correction)
    plot_correlation_matrix(
        phik_matrix.values,
        x_labels=phik_matrix.columns,
        y_labels=phik_matrix.index,
        vmin=0,
        vmax=1,
        color_map="Blues",
        title=r"correlation $\phi_K$",
        fontsize_factor=1.5,
        figsize=(7, 5.5),
        pdf_file_name=plot_file_name,
    )
    if pdf_file_name:
        plt.savefig(pdf_file, format="pdf", bbox_inches="tight", pad_inches=0)
        plt.show()

    ### 1b. global correlations
    if store_each_plot:
        plot_file_name = folder + "global_phik.pdf"
        output_files["global_phik"] = plot_file_name
    global_phik, global_labels = global_phik_from_rebinned_df(
        data_binned, noise_correction
    )
    plot_correlation_matrix(
        global_phik,
        x_labels=[""],
        y_labels=global_labels,
        vmin=0,
        vmax=1,
        figsize=(3.5, 4),
        color_map="Blues",
        title=r"$g_k$",
        fontsize_factor=1.5,
        pdf_file_name=plot_file_name,
    )
    # plt.tight_layout()
    if pdf_file_name:
        plt.savefig(pdf_file, format="pdf", bbox_inches="tight", pad_inches=0)
        plt.show()

    ### 2. Significance
    if store_each_plot:
        plot_file_name = folder + "significance_matrix.pdf"
        output_files["significance_matrix"] = plot_file_name
    significance_matrix = significance_from_rebinned_df(
        data_binned,
        lambda_significance,
        simulation_method,
        nsim_chi2,
        significance_method,
    )
    plot_correlation_matrix(
        significance_matrix.fillna(0).values,
        x_labels=significance_matrix.columns,
        y_labels=significance_matrix.index,
        vmin=-5,
        vmax=5,
        title="significance",
        usetex=False,
        fontsize_factor=1.5,
        figsize=(7, 5.5),
        pdf_file_name=plot_file_name,
    )
    if pdf_file_name:
        plt.savefig(pdf_file, format="pdf", bbox_inches="tight", pad_inches=0)
        plt.show()

    ### 3. Outlier significance
    outliers_overview = {}
    if do_outliers:
        for i, comb in enumerate(itertools.combinations(data_binned.columns, 2)):
            c0, c1 = comb
            if (
                abs(significance_matrix.loc[c0, c1]) < significance_threshold
                or phik_matrix.loc[c0, c1] < correlation_threshold
            ):
                continue

            zvalues_df = outlier_significance_matrix_from_rebinned_df(
                data_binned[[c0, c1]].copy(), binning_dict, CI_method=CI_method
            )

            combi = ":".join(comb).replace(" ", "_")
            xlabels = zvalues_df.columns
            ylabels = zvalues_df.index
            xlabel = zvalues_df.columns.name
            ylabel = zvalues_df.index.name

            if store_each_plot:
                plot_file_name = folder + "pulls_{0:s}.pdf".format(combi)
                output_files[combi] = plot_file_name

            plot_correlation_matrix(
                zvalues_df.values,
                x_labels=xlabels,
                y_labels=ylabels,
                x_label=xlabel,
                y_label=ylabel,
                vmin=-5,
                vmax=5,
                title="outlier significance",
                identity_layout=False,
                fontsize_factor=1.2,
                pdf_file_name=plot_file_name,
            )

            outliers_overview[combi] = zvalues_df

            if pdf_file_name:
                plt.savefig(pdf_file, format="pdf", bbox_inches="tight", pad_inches=0)
                plt.show()

    # save plots
    if pdf_file_name:
        output_files["all"] = pdf_file_name
        plt.close()
        pdf_file.close()

    return (
        phik_matrix,
        global_phik,
        significance_matrix,
        outliers_overview,
        output_files,
    )
